/*  
 * Copyright (c) 2011 Ant Kutschera
 * 
 * This file is part of Ant Kutschera's blog, 
 * http://blog.maxant.co.uk
 * 
 * This is free software: you can redistribute
 * it and/or modify it under the terms of the
 * Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 * 
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the Lesser GNU General Public License for
 * more details. 
 * 
 * You should have received a copy of the
 * Lesser GNU General Public License along with this software.
 * If not, see http://www.gnu.org/licenses/.
 */
package ch.maxant.comms.client.internal;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.net.URL;
import java.net.URLConnection;

import org.apache.commons.codec.binary.Base64;

import ch.maxant.comms.client.ClientConfiguration;
import ch.maxant.comms.client.RequestMarshaller;
import ch.maxant.comms.client.ResponseUnmarshaller;
import ch.maxant.comms.client.ServiceFactory;
import ch.maxant.comms.client.ServiceNameResolver;

public class ServiceFactoryImpl implements ServiceFactory {

	/**
	 * creates a proxy to the given service which can be used to call the remote system.
	 * @param <T> The service type
	 * @param serviceInterface the service interface
	 * @return the service proxy, which clients can use to call the server.
	 */
	@SuppressWarnings("unchecked")
	@Override
	public <T> T createServiceProxy(Class<T> serviceInterface, ClientConfiguration config){
		
		ClassLoader classLoader = serviceInterface.getClassLoader();

		Class<?>[] interfaces = new Class[1];
		interfaces[0] = serviceInterface;
		
		Handler h = new Handler(serviceInterface, config);

		return (T) Proxy.newProxyInstance(classLoader, interfaces, h);
	}

	private class Handler implements InvocationHandler {
		
		private Class<?> serviceInterface;
		private final ClientConfiguration config;
		
		public Handler(Class<?> serviceInterface, ClientConfiguration config) {
			this.serviceInterface = serviceInterface;
			this.config = config;
		}
		
		@Override
		public Object invoke(Object proxy, Method method, Object[] args)
		throws Throwable {

			//TODO do we need to use any kind of url encoding? prolly not, as this is a post body...
			
			////////////////////
			//CREATE MARSHALLERS & HELPERS
			////////////////////
			
			Class<?> marshallerClass = Class.forName(config.getMarshallerClass());
			RequestMarshaller marshaller = (RequestMarshaller) marshallerClass.newInstance();

			Class<?> unmarshallerClass = Class.forName(config.getUnmarshallerClass());
			ResponseUnmarshaller unmarshaller = (ResponseUnmarshaller) unmarshallerClass.newInstance();

			Class<?> serviceNameResolverClass = Class.forName(config.getServiceNameResolverClass());
			ServiceNameResolver nameResolver = (ServiceNameResolver) serviceNameResolverClass.newInstance();
			
			////////////////////
			//CREATE PAYLOAD
			////////////////////
			
			String payload = marshaller.marshal(config, serviceInterface, method, args, nameResolver);
			
			URL url = new URL(config.isSecure()?"HTTPS":"HTTP", config.getHost(), config.getPort(), config.getPath());
			URLConnection uc = url.openConnection();
			uc.setDoOutput(true);
			if(config.getUsername() != null){
				String encodedCredentials = new String(Base64.encodeBase64((config.getUsername() + ":" + config.getPassword()).getBytes()));
				uc.setRequestProperty("Authorization", "Basic " + encodedCredentials);
			}
			uc.setRequestProperty("IS_ZIPPING", String.valueOf(config.isUseZipping()));  //TODO could to this encoding, so that server unzips itself? 
			uc.setRequestProperty("Content-type", "text/html; charset=" + config.getCharEncoding());
			
			//TODO if required from config, do zipping
			//TODO ignore certificate authoriation if required?

			////////////////////
			//WRITE REQUEST
			////////////////////

			PrintWriter pw = null;
			try {
				pw = new PrintWriter(uc.getOutputStream());
				pw.print(payload);
			} finally {
				if(pw != null){
					pw.close(); 
				}
			}

			////////////////////
			//READ RESPONSE
			////////////////////
			BufferedReader in = null;
			StringBuilder response = new StringBuilder();
			try{
				in = new BufferedReader(new InputStreamReader(uc.getInputStream()));
				String current = null;
				while((current = in.readLine()) != null){
					response.append(current);
				}
			}finally{
				if(in != null){
					in.close();
				}
			}
			
			return unmarshaller.unmarshal(config, response.toString());
		}
	}
}
