/*  
 * Copyright (c) 2011 Ant Kutschera
 * 
 * This file is part of Ant Kutschera's blog, 
 * http://blog.maxant.co.uk
 * 
 * This is free software: you can redistribute
 * it and/or modify it under the terms of the
 * Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 * 
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the Lesser GNU General Public License for
 * more details. 
 * 
 * You should have received a copy of the
 * Lesser GNU General Public License along with this software.
 * If not, see http://www.gnu.org/licenses/.
 */
package ch.maxant.comms.server.servlet;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.UndeclaredThrowableException;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import ch.maxant.comms.server.Request;
import ch.maxant.comms.server.RequestUnmarshaller;
import ch.maxant.comms.server.ResponseMarshaller;
import ch.maxant.comms.server.ServiceLocator;
import ch.maxant.comms.server.internal.DefaultRequestUnmarshaller;
import ch.maxant.comms.server.internal.DefaultResponseMarshaller;
import ch.maxant.comms.server.internal.DefaultServiceLocator;

/**
 * this servlet listens for post requests and converts the posted message to a server call.
 */
public class CommsEndpoint extends HttpServlet {
	
	private static final long serialVersionUID = 1L;

	public static final String CH_MAXANT_COMMS_REQUEST_UNMARSHALLER = "ch.maxant.comms.request_unmarshaller";
	public static final String CH_MAXANT_COMMS_RESPONSE_MARSHALLER = "ch.maxant.comms.response_marshaller";
	public static final String CH_MAXANT_COMMS_SERVICE_LOCATOR = "ch.maxant.comms.service_locator";
	
	@Override
	protected void doGet(HttpServletRequest request, HttpServletResponse response)
			throws ServletException, IOException {
		response.getWriter().write("you have found the end point!  use a post request to call a service!");
	}
	
	/** {@inheritDoc} */
	@Override
	protected void doPost(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
		
		try{
			//TODO handle zipping
			//boolean isZippingHeader = Boolean.parseBoolean(request.getHeader("IS_ZIPPING"));
	
			///////////////////////////////
			// create unmarshaller
			///////////////////////////////
			String requestUnmarshallerClass = request.getSession().getServletContext().getInitParameter(CH_MAXANT_COMMS_REQUEST_UNMARSHALLER);
			if(requestUnmarshallerClass == null){
				requestUnmarshallerClass = DefaultRequestUnmarshaller.class.getName();
			}
			//TODO cache of services and methods here - for performance, but weak refs
			Class<?> c = Class.forName(requestUnmarshallerClass);
			RequestUnmarshaller unmarshaller = (RequestUnmarshaller) c.newInstance();
			Request req = unmarshaller.unmarshal(request);
	
			///////////////////////////////
			// create marshaller
			///////////////////////////////
			String responseMarshallerClass = request.getSession().getServletContext().getInitParameter(CH_MAXANT_COMMS_RESPONSE_MARSHALLER);
			if(responseMarshallerClass == null){
				responseMarshallerClass = DefaultResponseMarshaller.class.getName();
			}
			//TODO cache of services and methods here - for performance, but weak refs
			c = Class.forName(responseMarshallerClass);
			ResponseMarshaller marshaller = (ResponseMarshaller) c.newInstance();
			
			//now we have all the info we need, lets start unmarshalling and call the service
			Object service = lookupService(request, req);
			
			String resp = makeServiceCall(req, service, request, marshaller);
			
			response.getWriter().write(resp == null?"":resp);

		}catch(ClassNotFoundException e){
			String msg = "failed to find class while handling a service call: " + e.getMessage();
			log(msg, e);
			response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, msg);
		} catch (InstantiationException e) {
			String msg = "failed to instantiate something while handling a service call: " + e.getMessage();
			log(msg, e);
			response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, msg);
		} catch (IllegalAccessException e) {
			String msg = "illegal access on something while handling a service call: " + e.getMessage();
			log(msg, e);
			response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, msg);
		} catch (NoSuchMethodException e) {
			String msg = "no method found while handling a service call: " + e.getMessage();
			log(msg, e);
			response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, msg);
		} catch (InvocationTargetException e) {
			String msg = "failed to call a proxy while handling a service call: " + e.getMessage();
			log(msg, e);
			response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, msg);
		} catch (Exception e) {
			String msg = "something bad went wrong while handling a service call: " + e.getMessage();
			log(msg, e);
			response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, msg);
		}
	}

	private String makeServiceCall(Request req, Object service, HttpServletRequest request, ResponseMarshaller marshaller)
			throws NoSuchMethodException, IllegalAccessException,
			InvocationTargetException, IOException {
		//TODO cache of services and methods here - for performance, but weak refs
		Class<?> serviceClass = service.getClass();
		Method serviceMethod = null;
		serviceMethod = serviceClass.getMethod(req.getMethodName(), req.getParameterTypes().toArray(new Class<?>[0]));
		if(serviceMethod == null){
			serviceMethod = serviceClass.getDeclaredMethod(req.getMethodName(), req.getParameterTypes().toArray(new Class<?>[0]));
			if(serviceMethod == null){
				//TODO search in super class?
				throw new NoSuchMethodException(req.getMethodName());
			}
		}
		
		Object ret = null;
		
		try{
			ret = serviceMethod.invoke(service, req.getArguments().toArray());
		}catch(InvocationTargetException e){
			//checks the exceptions which are declared. if its one of them, then the client must know it,
			//otherwise, just throw a {@link RuntimeException} with suitable text.
			for(Class<?> c : serviceMethod.getExceptionTypes()){
				if(c.isAssignableFrom(e.getTargetException().getClass())){
					ret = e.getTargetException();
					break;
				}
			}
			
			if(ret == null){
				//none found
				//throw a suitable exception, which wont cause the client to crash.
				//imagine if we throw a type that the client didnt have the class for!
				//pass null, because we cannot guarantee that the client will have
				//this exception
				UndeclaredThrowableException ute = new UndeclaredThrowableException(new Exception("The server threw a " + e.getClass() + " exception which was not declared.  Error was: " + e.getMessage()));
				ute.setStackTrace(e.getTargetException().getStackTrace());
				ret = ute;
			}
		}
		String resp = marshaller.marshal(req.getSerializationType(), ret, request);

		return resp;
	}


	private Object lookupService(HttpServletRequest request, Request req) throws ClassNotFoundException, InstantiationException, IllegalAccessException {
		String serviceLocatorClass = request.getSession().getServletContext().getInitParameter(CH_MAXANT_COMMS_SERVICE_LOCATOR);
		if(serviceLocatorClass == null){
			serviceLocatorClass = DefaultServiceLocator.class.getName();
		}
		Class<?> c = Class.forName(serviceLocatorClass);
		ServiceLocator sl = (ServiceLocator) c.newInstance();
		Object service = sl.lookup(req.getServiceName());
		return service;
	}

}
